# == MODULE region.py ==

"""
Genomic region module.
'Full' version, not to be shared with the participants.
:author: András Aszódi
:date: 2020-11-10
"""

import re

class Region:
    """
    Instances of the Region class represent genomic regions.
    A Region object stores a chromosome, start and end coordinates,
    an arbitrary region name, and a strand direction indicator.
    It can be converted to and from strings formatted as BED or GFF.
    """

    # 2. The initialiser
    def __init__(self, chrom="chrUn",
                 start=0, end=0,
                 name=".", score=0.0, strand="."):
        """
        Creates a new Region object.
        :param chrom: The chromosome name, default "chrUn" (by convention this means "unknown")
        :param start: The start coordinate. Always 0-based ("BED convention")
        :param end: The end coordinate. Half-open, BED convention.
            If `end` < `start` then they are swapped silently.
        :param name: The name of the region, can be an arbitrary string
        :param score: Some number, e.g. in BED files it is between 0 and 1000
        :param strand: Strand direction indicator, can be "+","-", or "." (the default).
            Only the first character of the parameter is used.
            Any other character will be converted to "." silently.
        """
        self.chrom = chrom
        if start > end:
            end, start = start, end
        self.start = start
        self.end = end
        self.name = name
        self.score = score
        s0 = strand[0]
        self.strand = s0 if s0 in "+-." else "."

    # -- Conversion from/to strings --

    # 3. Input
    def from_bed(self, line):
        """
        Parses a string according to the BED standard.
        If parsing is successful, update the calling object with the new values.
        :param line: A string, possibly a line read from a BED file.
        :return: True if parsing was successful and the calling object was modified,
            False if `line` was empty/all-whitespace/comment and the calling object was not changed.
        :raise: ValueError if parsing was unsuccessful
        """
        # 3A-1. Ignore "empty" lines
        if self._ignore_line(line):
            return False

        # 3B. Split line in fields
        fields = line.split("\t")
        flen = len(fields)
        if flen < 3:
            raise ValueError("Too few fields")

        # 3C-1. Parse input fields
        chrom = fields[0]
        start = int(fields[1])  # may raise ValueError
        end = int(fields[2])
        name = fields[3] if flen >= 4 else "."
        score = float(fields[4]) if flen >= 5 else 0.0
        strand = fields[5][0] if flen >= 6 else "."

        # 3C-2. Update calling object
        self.chrom = chrom
        # automatic swapping if start > end
        self.start, self.end = (start, end) if start <= end else (end, start)
        self.name = name
        self.score = score
        self.strand = strand
        return True

    # HW
    def from_gff(self, line):
        """
        Parses a string according to the GFF standard.
        If parsing is successful, update the calling object with the new values.
        :param line: A string, possibly a line read from a GFF file.
        :return: True if parsing was successful and the calling object was modified,
            False if `line` was empty/all-whitespace/comment and the calling object was not changed.
        :raise: ValueError if parsing was unsuccessful
        """
        # Implementation is left as an exercise to the participants
        ### **TEACHER'S NOTE** : the "skel" version should contain a `pass` here
        # pass
        ### **TEACHER'S NOTE** : implementation
        # Ignore "empty" lines
        if self._ignore_line(line):
            return False

        # Split line in fields
        fields = line.split("\t")
        flen = len(fields)
        if flen < 5:
            raise ValueError("Too few fields")

        # Parse input fields
        chrom = fields[0]
        src = fields[1]  # "source" field: not stored
        name = fields[2]    # "feature" field
        start = int(fields[3]) - 1 # may raise ValueError: note silly GFF start offset convention
        end = int(fields[4])
        score = float(fields[5]) if flen >= 6 else 0.0
        strand = fields[6][0] if flen >= 7 else "."
        # ... ignore "frame" and "group" fields

        # Update calling object
        # Same as `from_bed()`, just repeated here
        self.chrom = chrom
        # automatic swapping if start > end
        self.start, self.end = (start, end) if start <= end else (end, start)
        self.name = name
        self.score = score
        self.strand = strand
        return True
        ### **END**

    # 4. Output
    def to_gff(self, src="genreg"):
        """
        Converts a Region object to its GFF string representation.
        :param src: This string will be used as the value of the GFF source field
        :returns: The string representation according to the GFF specification.
        """
        return "\t".join( (
            self.chrom,
            src,     # the GFF source field
            self.name,
            str(self.start + 1),  # BED->GFF coordinate transformation
            str(self.end),
            str(self.score),
            self.strand,
            ".", ".",    # "frame" and "group" columns that we do not worry about
        ) )

    # HW
    def to_bed(self):
        """
        Converts a Region object to its BED string representation.
        :returns: The string representation according to the GFF specification.
        """
        # Implementation is left as an exercise to the participants
        ### **TEACHER'S NOTE** : the "skel" version should contain a `pass` here
        # pass
        ### **TEACHER'S NOTE** : implementation
        return "\t".join( (
            self.chrom,
            str(self.start),
            str(self.end),
            self.name,
            str(self.score),
            self.strand
        ) )

    # Examples of "special" methods
    # They always have two underscores at the beginning and end of their names.
    # The two methods below implement equality comparison for the Region class.
    # In our case it is really simple but in general you may have special requirements,
    # for instance when string members have to be compared in a case-insensitive manner.

    # 5. Equality
    def __eq__(self, other):
        """
        Two Regions are considered equal if all their attributes are equal.
        This method will be invoked in expressions like `if region1 == region2: ...`
        """

        # Use this idiom if all members are supposed to be the same.
        # The __dict__ special member lists all members of a class (or object)
        # in a dictionary.
        return type(other) is type(self) and \
               self.__dict__ == other.__dict__

    # Properties
    # You can invoke methods that have been decorated with `@property`
    # without the parentheses, imitating member access.

    # 6. Region size
    @property
    def size(self):
        """
        :return: The size (length) of the region.
        """
        return self.end - self.start

    # -- "private" methods --

    # There are no truly private (hidden) methods in Python.
    # Encapsulation (hiding of implementation details) can be emulated
    # by a naming convention: any data member or method is considered "private'
    # by convention if its name starts by an underscore.

    # 3A-2. Detects "empty" lines to be ignored by the input methods.
    def _ignore_line(self, line):
        """
        :param line: A string containing a line, possibly read from a BED or GFF file.
        :return: True if `line` is empty or all-whitespace or a comment.
        """
        return len(line) == 0 or \
            re.match(r"[ \t]*$", line) or \
            line[0] == '#'




# == UNIT TEST testregion.py ==

"""
Unit tests for the region.Region class.
"""

import unittest

# 1A. Import the Region class
from ..region import Region

class TestRegion(unittest.TestCase):

    # -- Setting the stage --
    
    # The setUpClass method is invoked only once at the beginning of the testing.
    # Put initialisation code here that applies to all tests.
    @classmethod
    def setUpClass(cls):
        # The correctly formatted strings are constant so we can set them up only once
        # as class variables.
        cls.goodbed = "chr1\t15\t25\tgood\t7.2\t+"
        cls.goodgff = "chr1\tmysrc\tgood\t16\t25\t7.2\t+\t.\t."

    # 1B. The setUp method is invoked before each testing method.
    # Use it to create objects/variables "afresh", open files etc.
    def setUp(self):
        # create an empty region
        self.empty = Region()
        
        # create a good region
        self.r = Region("chr1", 15, 25, "good", 7.2, "+")
    
    # -- The testing methods --
    
    # The names of these methods must begin with `test`.
    # Each method should test only one particular functionality.
    # To compare observed and expected values, make use of the 
    # various `assertXXXX` methods of the `unittest.TestCase` class.
    # The test methods are invoked automatically by the 
    # `main` method (see below). The order of invocation is undefined.
    # NEVER write test methods that somehow depend on each other!

    # 2. Test __init__
    def test_init(self):
        # 2A. Properly initialised region
        self.assertEqual(self.r.chrom, "chr1")
        self.assertEqual(self.r.start, 15)
        self.assertEqual(self.r.end, 25)
        self.assertEqual(self.r.name, "good")
        self.assertEqual(self.r.score, 7.2)
        self.assertEqual(self.r.strand, "+")
        
        # 2B. Automatic corrections
        r = Region("chr2", 30, 15, "swap", strand = "X")
        self.assertEqual(r.start, 15)
        self.assertEqual(r.end, 30)
        self.assertEqual(r.strand, ".")

        ### Additional assignment: test an empty Region
        self.assertEqual(self.empty.chrom, "chrUn")
        self.assertEqual(self.empty.start, 0)
        self.assertEqual(self.empty.end, 0)
        self.assertEqual(self.empty.name, ".")
        self.assertEqual(self.empty.score, 0.0)
        self.assertEqual(self.empty.strand, ".")

    # 3. Test reading from a BED-formatted string
    def test_from_bed(self):
        # 3A. Empty string, all-whitespace string, comment
        for line in ("", " \t ", "# comment"):
            self.assertFalse(self.r.from_bed(line))

        # 3B. Check number of fields
        with self.assertRaises(ValueError, msg="Too few fields"):
            self.r.from_bed("chr\t123")

        # 3C. Correct format
        ok = self.empty.from_bed(self.goodbed)
        self.assertTrue(ok)
        ### Additional assignment
        # self.empty must be equal to self.r now (the same data were read)
        self.assertEqual(self.empty, self.r)

        ### Additional assignment
        # check invalid integer fields
        with self.assertRaises(ValueError, msg="invalid literal for int() with base 10: 'joe'"):
            self.r.from_bed("chr\t1\tjoe")

    # 4. Test writing GFF-formatting string
    def test_to_gff(self):
        gffstr = self.r.to_gff(src="mysrc")
        self.assertEqual(gffstr, self.goodgff)

    # 5. Test equality
    def test_equal(self):
        """test (in)equality of two Region objects"""
        er = Region("chr1", 15, 25, "good", 7.2, "+")
        self.assertEqual(self.r, er)
        dr = Region("chr1", 15, 25, "good", 3.3, "-")
        self.assertNotEqual(self.r, dr)

    # 6. Test region size
    def test_size(self):
        self.assertEqual(self.r.size, 10)

    # *** Homework tests ***

    # NOTE: these tests are _NOT_ copied into the "skel" version of `testregion.py`

    # X3. Test reading from a GFF-formatted string
    def test_from_gff(self):
        # X3A. Empty string, all-whitespace string, comment
        for line in ("", " \t ", "# comment", "##gff_version 2"):
            self.assertFalse(self.r.from_gff(line))

        # X3B. Check number of fields
        with self.assertRaises(ValueError, msg="Too few fields"):
            self.r.from_gff("chr\tsrc\t123")

        # X3C. Correct format
        ok = self.empty.from_gff(self.goodgff)
        self.assertTrue(ok)
        ### Additional assignment
        # self.empty must be equal to self.r now (the same data were read)
        self.assertEqual(self.empty, self.r)

        ### Additional assignment
        # check invalid integer fields
        with self.assertRaises(ValueError, msg="invalid literal for int() with base 10: 'joe'"):
            self.r.from_gff("chr\tsrc\tbadregion\t1\tjoe")

    # X4. Test writing BED-formatting string
    def test_to_bed(self):
        bedstr = self.r.to_bed()
        self.assertEqual(bedstr, self.goodbed)

    # *** End of homework tests ***

    # -- After testing --
    
    # 1C. The tearDown method is invoked after each testing method has been run.
    # Use it to clean up the remnants of tests, close files.
    # It is used less often than its pendant setUp .
    def tearDown(self):
        pass

    # The tearDownClass method is invoked after the whole testing has finished.
    # It can be used to clean up things that were set up by setUpClass
    # at the very beginning.
    @classmethod
    def tearDownClass(cls):
        pass


# 2C. Run the unit tests
if __name__ == "__main__":
    unittest.main()

